/**************************************************************************************
Author:         David Szybala
Website:        www.davidszybala.com
Alt Website:    www.twistedgerbil.com
Date:           11/19/2014
Program:        material_list
Description:    This script allows the user to perform special functions with materials
                (assignments, selecting by material, showing in editor, etc) 
***************************************************************************************/

try (material_list_form.close()) catch()


struct MaterialList
(
    materials_root = #(),
    materials_all = #(),
    materials_offset = #(),
    materials_filtered = #(),
    materials_offset_filtered = #(),
    material_children = #(),
    
    /**************************************************************************
    Sorts items by name
    **************************************************************************/
    fn sort_by_name item_one item_two =
    (
        case of
        (
            (item_one.name < item_two.name): -1
            (item_one.name > item_two.name): 1
            default: 0
        ) --end case
    ), --end fn
    
    /**************************************************************************
    Saves an empty file then deletes it. This is needed to force max to properly
    update its internal 'sceneMaterials' variable
    **************************************************************************/
    fn force_material_update =
    (
        saveNodes #() (GetDir #scene + "\\_MLU_TEMP.max") quiet:true
        deleteFile (GetDir #scene + "\\_MLU_TEMP.max") 
    ), --end fn
    
    /**************************************************************************
    Returns all root materials in the scene sorted by name
    **************************************************************************/
    fn get_root_materials =
    (
        materials = for current in sceneMaterials where (superClassOf current) == material collect current
        
        if materials.count > 0 do
            qsort materials sort_by_name
        
        materials --return value
    ), --end fn
    
    /**************************************************************************
    Removes all root materials that do not have a valid dependent. Valid
    dependents are geometry or the renderer. This is the poor mans way of
    removing unused materials
    **************************************************************************/
    fn remove_unused_materials material_list =
    (
        for i = material_list.count to 1 by -1 do
        (
            if (refs.dependentNodes material_list[i] firstOnly:true) == undefined do
            (
                dependents = for current in (refs.dependents material_list[i]) collect (superClassOf current)
                
                if (findItem dependents RendererClass) == 0 and (findItem dependents atmospheric) == 0 do
                    deleteItem material_list i
            ) --end if
        ) --end for
        
        material_list --return value
    ), --end fn
    
    /**************************************************************************
    Recursive function. Collects all the sub materials from the argument
    'parent_material' WARNING: Some material (eg VRayBlendMtl) will report the
    same sub material multiple times. Check each sub material against a master
    list to prevent repeats
    **************************************************************************/
    fn collect_sub_materials parent_material offset_value =
    (
        local name_offset = ""
        
        for x = 1 to (getNumSubMtls parent_material) do
        (
            child_material = getSubMtl parent_material x
            
            if child_material != undefined and (findItem this.material_children child_material) == 0 do
            (
                append this.materials_all child_material
                append this.material_children child_material
                
                for y = 1 to offset_value do
                    name_offset += "   "
                
                append this.materials_offset name_offset
                
                --CRITICAL! You must reset 'name_offset' for recursion
                name_offset = ""
                
                if (getNumSubMtls child_material) > 0 do
                    collect_sub_materials child_material (offset_value + 1)
            ) --end if
        ) --end for
    ), --end fn
    
    /**************************************************************************
    Filters the current scene materials by name. NOTE: Sub materials are NOT affected
    **************************************************************************/
    fn filter_scene_materials filter_text =
    (
        local append_sub_materials = true
        
        --CRITICAL! Clear previous data
        free this.materials_filtered
        free this.materials_offset_filtered
        
        if filter_text != "" then
        (
            for i = 1 to this.materials_offset.count do
            (
                if this.materials_offset[i] == "" then
                (
                    if (matchPattern this.materials_all[i].name pattern:("*" + filter_text + "*") ignorecase:false) then
                    (
                        append this.materials_filtered this.materials_all[i]
                        append this.materials_offset_filtered this.materials_offset[i]
                        append_sub_materials = true
                    ) --end if
                    else
                        append_sub_materials = false
                ) --end if
                else
                (
                    if append_sub_materials do
                    (
                        append this.materials_filtered this.materials_all[i]
                        append this.materials_offset_filtered this.materials_offset[i]
                    ) --end if
                ) --end else
            ) --end for
        ) --end if
        else
        (
            this.materials_filtered = deepcopy this.materials_all
            this.materials_offset_filtered = deepcopy this.materials_offset
        ) --end else
    ), --end fn
    
    /**************************************************************************
    Collects a list of all requested materials current scene (root only or plus subs).
    Also applies filtering to the list
    **************************************************************************/
    fn collect_scene_materials force_update include_subs filter_text =
    (
        --CRITICAL! Clear previous data
        free this.materials_all
        free this.materials_offset
        
        if force_update then
        (
            this.force_material_update()
            this.materials_root = this.get_root_materials()
        ) --end if
        else
        (
            this.materials_root = this.get_root_materials()
            this.materials_root = this.remove_unused_materials this.materials_root
        ) --end else
        
        for current in this.materials_root do
        (
            append this.materials_all current
            append this.materials_offset ""
        
            if include_subs do
            (
                if (getNumSubMtls current) > 0 do
                (
                    --CRITICAL! Clear previous data!
                    free this.material_children
                    
                    collect_sub_materials current 1
                ) --end if
            ) --end if
        ) --end for
        
        this.filter_scene_materials filter_text
    ), --end fn
    
    /**************************************************************************
    Places the materials defined by 'index_list' in the material editor
    starting at slot 'slot'
    **************************************************************************/
    fn set_editor_slot_by_index index_list slot =
    (
        for index in index_list while slot <= 24 do
        (
            setMeditMaterial slot this.materials_filtered[index]
            slot += 1
        ) --end for
    ), --end fn
    
    /**************************************************************************
    Places the materials defined by 'index_list' in the slate material editor.
    Returns true on success
    **************************************************************************/
    fn set_slate_editor_materials index_list =
    (
        local success = true
        
        --CRITICAL! The slate editor must be open at this point!
        SME.Open()
        sme_view = SME.getView (SME.activeView)

        if sme_view == undefined do
        (
            sme_view = SME.createView "View1"
            sme_view = SME.getView (SME.activeView)
        ) --end if

        for i = 1 to index_list.count do
        (
            node_postition = [0, (i * 100)]
            sme_view.createNode this.materials_filtered[index_list[i]] &node_postition
        ) --end for

        try
        (
            --sme_view.zoomExtents type:#all
            hwnd = windows.getChildHWND 0 "Slate Material Editor"
            WM_COMMAND = 0x111
            ID = 0xA02C -- Layout All Button source ID
            windows.sendMessage hwnd[1] WM_COMMAND ID 0 -- press button
        ) --end try
        catch
        (
            success = false
        ) --enc catch
        
        success --return value
    ), --end fn
    
    /**************************************************************************
    Returns the root material index for the argument 'sub_index'
    **************************************************************************/
    fn get_root_index_for_material sub_index =
    (
        local root_index = sub_index
        
        while this.materials_offset[root_index][1] == " " do
            root_index -= 1
        
        root_index --return value
    ), --end fn
    
    /**************************************************************************
    Returns a list of objects assigned materials in argument 'index_list'
    **************************************************************************/
    fn get_objects_with_material index_list =
    (
        local found_objects = #()
        
        for current in objects do
        (
            for index in index_list do
            (
                if current.material == this.materials_filtered[index] do
                    append found_objects current
            ) --end for
        ) --end for
        
        found_objects --return value
    ), --end fn
    
    /**************************************************************************
    Assigns the material at index 'index' to the current selection
    **************************************************************************/
    fn assign_material_to_selection index =
    (
        for current in selection do
            current.material = this.materials_filtered[index]
    ), --end fn
    
    /**************************************************************************
    Returns a list of the materials assigned to the selected objects
    **************************************************************************/
    fn get_materials_from_selected =
    (
        local selection_materials = #()
        
        for current in selection do
        (
            if current.material != undefined do
                appendIfUnique selection_materials current.material
        ) --end for
        
        selection_materials --return value
    ), --end fn
    
    /**************************************************************************
    Places the materials in 'material_list' in the material editor
    starting at slot 'slot'
    **************************************************************************/
    fn set_editor_slot_by_material material_list slot =
    (
        for current in material_list while slot <= 24 do
        (
            setMeditMaterial slot current
            slot += 1
        ) --end for
    ), --end fn
    
    /**************************************************************************
    Assigns the id value 'id' to all objects in 'object_list'
    **************************************************************************/
    fn assign_id_to_objects object_list id =
    (
        for current in object_list do
            current.gbufferchannel = id
    ), --end fn
    
    /**************************************************************************
    Assigns the id value 'id' to all materials defined by 'index_list'
    **************************************************************************/
    fn assign_id_to_materials index_list id=
    (
        for index in index_list do
            this.materials_filtered[index].effectsChannel = id
    ), --end fn
    
    /**************************************************************************
    Instances all root materials with the same name. Returns the number of
    materials replaced
    **************************************************************************/
    fn instance_duplicate_materials =
    (
        local current_name = ""
        local counter = 0
        
        for i = 1 to (this.materials_root.count - 1) do
        (
            current_name = this.materials_root[i].name
            
            for x = i + 1 to this.materials_root.count do
            (
                if this.materials_root[x].name == current_name do
                (
                    replaceInstances this.materials_root[x] this.materials_root[i]
                    counter += 1
                ) --end if
            ) --end for
        ) --end for
        
        counter --return value
    ), --end fn
    
    /**************************************************************************
    Saves the indexed materials to file 'file_path' Returns true on success
    **************************************************************************/
    fn save_materials index_list file_path =
    (
        local material_library = materialLibrary()
        
        for index in index_list do
            append material_library this.materials_filtered[index]
        
        saveTempMaterialLibrary material_library file_path --return value
    ) --end fn
) --end struct


( --local scope
    global material_list_form
    local material_list_ini = getDir #plugcfg + "\\MaterialList.ini"

    --CRITICAL! These controls/variables need to be accessed in event handlers!
    local mat_list = MaterialList()

    local materials_group = undefined
    local material_listview = undefined
    local listbox_cache = undefined
    local material_filter_text = undefined
    local show_sub_checkbox = undefined
    local force_update_checkbox = undefined
    local slot_spinner = undefined
    local defalut_material_checkbox = undefined
    local add_selection_checkbox = undefined
    local obj_id_spinner = undefined
    local mat_id_spinner = undefined
    local status_label = undefined

    /********************************************************************
    Form closing event handler. Stores form information for next time
    ********************************************************************/
    fn material_list_form_closing sender events =
    (
        local form_location = material_list_form.location
        local form_size = material_list_form.size

        form_location = point2 form_location.x form_location.y
        form_size = point2 form_size.width form_size.height

        setIniSetting material_list_ini "Dialog" "Position" (form_location as string)
        setIniSetting material_list_ini "Dialog" "Size" (form_size as string)

        material_list_form.Dispose()
        material_list_form = undefined
    ) --end fn

    /***********************************************************************
    Refreshes the listview control
    ***********************************************************************/
    fn refresh_listview =
    (
        --CRITICAL! "Reset" listview to avoid display bug!
        if material_listview.Items.Count > 0 do
        (
            material_listview.EnsureVisible(0)
            material_listview.SelectedIndices.Clear()
            material_listview.VirtualListSize = 0
        ) --end if

        material_listview.VirtualListSize = mat_list.materials_filtered.count
        listbox_cache = for i = 1 to mat_list.materials_filtered.count collect undefined
        material_listview.Refresh()
    ) --end fn

    /***********************************************************************
    Refreshes the material names in the list box
    ***********************************************************************/
    fn refresh_material_list =
    (
        mat_list.collect_scene_materials force_update_checkbox.checked show_sub_checkbox.checked material_filter_text.Text

        message = "Materials: " + (mat_list.materials_root.count as string)
        materials_group.text = message

        refresh_listview()
    ) --end fn

    /********************************************************************
    List resize event handler. Ensures the last column will fill the
    remaining space in the list
    ********************************************************************/
    fn material_listview_resized sender events =
    (
        (material_listview.Columns.Item 1).width = -2
    ) --end fn

    /********************************************************************
    Material virtual item event handler. Retrieves the requested item
    for the material listview
    ********************************************************************/
    fn material_listview_retrieve_virtual_item sender events =
    (
        local index = events.ItemIndex + 1

        if listbox_cache[index] == undefined then
        (
            current_item = dotNetObject "System.Windows.Forms.ListViewItem" ""
            current_item.Text = mat_list.materials_offset_filtered[index] + mat_list.materials_filtered[index].name
            current_item.SubItems.Add ((classOf mat_list.materials_filtered[index]) as string)

            sys_color = (colorMan.getColor #text) * 255
            current_item.ForeColor = (dotNetClass "System.Drawing.Color").FromArgb 255 sys_color[1] sys_color[2] sys_color[3]

            listbox_cache[index]  = current_item
            events.Item = current_item
        ) --end if
        else
            events.Item = listbox_cache[index]
    ) --end fn

    /***********************************************************************
    Filter changed event handler. Refreshes the material list based on the
    user filter
    ***********************************************************************/
    fn material_filter_text_changed sender events =
    (
        mat_list.filter_scene_materials material_filter_text.Text

        refresh_listview()
    ) --end fn

    /***********************************************************************
    Update button clicked event handler. Updates the UI
    ***********************************************************************/
    fn update_list_button_pressed sender events =
    (
        refresh_material_list()
        status_label.text = "INFO: Material list updated"
    ) --end fn

    /***********************************************************************
    Show sub material checkbox clicked event handler. Updates the UI to
    show/hide sub materials
    ***********************************************************************/
    fn show_sub_checkbox_changed sender events =
    (
        refresh_material_list()

        if show_sub_checkbox.Checked then
            status_label.text = "INFO: Showing sub materials"
        else
            status_label.text = "INFO: Hiding sub materials"
    ) --end fn

    /***********************************************************************
    Show material in editor button clicked event handler. Shows the selected
    material(s) in the editor starting at slot number <spinner value>
    ***********************************************************************/
    fn show_material_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)
        local editor_slot = slot_spinner.value
        local result = true
        local message = ""

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        if MatEditor.mode == #basic then
        (
            mat_list.set_editor_slot_by_index user_selection editor_slot

            if user_selection.count + editor_slot >= 25 do
                result = false
        ) --end if
        else
            result = mat_list.set_slate_editor_materials user_selection

        if result then
        (
            message = "INFO: " + (user_selection.count as string)
            message += " material(s) placed in editor"
        ) --end if
        else
        (
            if MatEditor.mode == #basic then
                message = "WARNING: Not enough slots available for selected materials"
            else
                message = "ERROR: Could not show material in slate editor"
        ) --end else

        status_label.text = message

        if not MatEditor.isOpen() do
            MatEditor.Open()
    ) --end fn

    /***********************************************************************
    Spinner clicked event handler. Emulates the right click to reset
    spinner functionality of 3ds max
    ***********************************************************************/
    fn spinner_mouse_click sender events =
    (
        if (events.Button == events.Button.Right) do
            sender.value = sender.Minimum
    ) --end fn

    /***********************************************************************
    Reset editor button clicked event handler. Sets all editor materials to
    their default state or deletes all slate editor views
    ***********************************************************************/
    fn reset_editor_button_pressed sender events =
    (
        local message = "INFO: Material editor reset"

        --Clean sample slots
        if defalut_material_checkbox.checked then
        (
            for i = 1 to 24 do
            (
                default_name = ((i as string) + " - Default")

                if i < 10 do
                    default_name = "0" + default_name

                meditmaterials[i] = StandardMaterial name:default_name
            ) --end for
        ) --end if
        else
            macros.run "Medit Tools" "clear_medit_slots"

        --Clean slate editor views if necessary
        if MatEditor.mode != #basic do
        (
            loop_control = true
            counter = 1

            while loop_control do
            (
                for i = 1 to SME.GetNumViews() do
                    SME.DeleteView i false

                if counter > 2 do
                    loop_control = false

                if SME.GetNumViews() > 0 then
                    counter += 1
                else
                    loop_control = false
            ) --end while

            if SME.GetNumViews() == 0 then
                SME.createView "View1"
            else
                message = "ERROR: Unable to reset slate editor views"
        ) --end if

        status_label.text = message
    ) --end fn

    /***********************************************************************
    Select objects with material button clicked event handler. Selects all
    objects with the selected materials assigned
    ***********************************************************************/
    fn select_objects_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)
        local root_selection = #()

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        for index in user_selection do
            appendIfUnique root_selection (mat_list.get_root_index_for_material index)

        found_objects = mat_list.get_objects_with_material root_selection

        if add_selection_checkbox.checked then
            selectMore found_objects
        else
            select found_objects

        message = "INFO: " + (found_objects.count as string) + " object(s) found"
        status_label.text = message
    ) --end fn

    /***********************************************************************
    Assign material to selection button clicked event handler. Assigns the
    selected material to the current selection
    ***********************************************************************/
    fn assign_material_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)
        local root_selection = #()

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        if user_selection.count > 1 do
        (
            status_label.text = "ERROR: More than one material selected"
            return false
        ) --end if

        if selection.count == 0 do
        (
            status_label.text = "ERROR: No objects selected"
            return false
        ) --end if

        for index in user_selection do
            appendIfUnique root_selection (mat_list.get_root_index_for_material index)

        mat_list.assign_material_to_selection root_selection[1]
        status_label.text = "INFO: Material assigned to selection"
    ) --end fn

    /***********************************************************************
    Show material of selected button clicked event handler. Shows the
    materials of the current selection in the editor
    ***********************************************************************/
    fn show_obj_mat_button_pressed sender events =
    (
        local selection_materials = mat_list.get_materials_from_selected()

        if selection_materials.count == 0 do
        (
            status_label.text = "ERROR: Selected objects do not have materials"
            return false
        ) --end if

        mat_list.set_editor_slot_by_material selection_materials 1

        if selection_materials.count > 24 then
            message = "WARNING: Not enough slots available for selected materials"
        else
            message = "INFO: Materials of selected objects placed in sample slots"

        status_label.text = message

        if not MatEditor.isOpen() do
            MatEditor.Open()
    ) --end fn

    /***********************************************************************
    Assign object id button clicked event handler. Assigns all objects
    with the selected materials an object id
    ***********************************************************************/
    fn obj_id_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)
        local root_selection = #()

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        for index in user_selection do
            appendIfUnique root_selection (mat_list.get_root_index_for_material index)

        found_objects = mat_list.get_objects_with_material root_selection

        mat_list.assign_id_to_objects found_objects obj_id_spinner.value

        message = "INFO: Object ID " + (obj_id_spinner.value as string)
        message += " assigned"
        status_label.text = message
    ) --end fn

    /***********************************************************************
    Assign material id button clicked event handler. Assigns all selected
    materials the defined material id
    ***********************************************************************/
    fn mat_id_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        mat_list.assign_id_to_materials user_selection mat_id_spinner.value

        message = "INFO: Material ID " + (mat_id_spinner.value as string)
        message += " assigned"
        status_label.text = message
    ) --end fn

    /***********************************************************************
    Save materials button clicked event handler. Saves the selected
    materials to a library file (*.mat)
    ***********************************************************************/
    fn save_materials_button_pressed sender events =
    (
        local user_selection = for i = 0 to (material_listview.SelectedIndices.Count - 1) collect ((material_listview.SelectedIndices.Item i) + 1)

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No materials selected"
            return false
        ) --end if

        local save_path = getSaveFileName caption:"Save Materials" types:("Material Library (*.mat)|*.mat")

        if save_path == undefined do
        (
            status_label.text = "WARNING: User cancelled material save"
            return false
        ) --end if

        result = mat_list.save_materials user_selection save_path

        if result then
            status_label.text = "INFO: Materials saved to file"
        else
            status_label.text = "ERROR: Could not save materials"
    ) --end fn

    /***********************************************************************
    Instance duplicate button clicked event handler. Instances all root
    materials with the same name
    ***********************************************************************/
    fn instance_duplicate_button_pressed sender events =
    (
        local replacements = mat_list.instance_duplicate_materials()
        refresh_material_list()

        message = "INFO: " + (replacements as string)
        message += " duplicate material(s) instanced"
        status_label.text = message
    ) --end fn

    /********************************************************************
    UI creation
    ********************************************************************/
    fn setup_ui =
    (
        material_list_form = dotNetObject "MaxCustomControls.MaxForm"
        material_list_form.name = "material_list_form"
        material_list_form.text = "Material List"
        material_list_form.Size = dotNetObject "System.Drawing.Size" 325 820
        dotNet.setLifetimeControl material_list_form #dotnet
        dotNet.addEventHandler material_list_form "FormClosing" material_list_form_closing
        material_list_form.suspendLayout()

        /********************************************************************
        Materials Group
        ********************************************************************/
        materials_group = dotNetObject "System.Windows.Forms.GroupBox"
        materials_group.name = "materials_group"
        materials_group.text = "Materials: 0"
        materials_group.Location = dotNetObject "System.Drawing.Point" 5 5
        materials_group.Size = dotNetObject "System.Drawing.Size" 300 305
        materials_group.flatStyle = materials_group.flatStyle.system
        materials_group.Anchor = dotnet.combineEnums materials_group.anchor.Top materials_group.anchor.Left materials_group.anchor.Right materials_group.anchor.Bottom
        materials_group.suspendLayout()

        material_list_form.controls.add materials_group

        material_listview = dotNetObject "System.Windows.Forms.ListView"
        material_listview.Name = "material_listview"
        material_listview.Location = dotNetObject "System.Drawing.Point" 10 20
        material_listview.Size = dotNetObject "System.Drawing.Size" 280 250
        material_listview.View = (dotNetClass "System.Windows.Forms.View").Details
        material_listview.FullRowSelect = true
        material_listview.HideSelection = false
        material_listview.Columns.Add "Name" 165
        material_listview.Columns.Add "Class" 115
        material_listview.Anchor = dotnet.combineEnums material_listview.anchor.Top material_listview.anchor.Left material_listview.anchor.Right material_listview.anchor.Bottom
        material_listview.VirtualMode = True
        material_listview.VirtualListSize = 0
        dotNet.setLifetimeControl material_listview #dotnet
        dotNet.addEventHandler material_listview "Resize" material_listview_resized
        dotNet.addEventHandler material_listview "RetrieveVirtualItem" material_listview_retrieve_virtual_item

        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #window) * 255
        material_listview.BackColor = dn_color current_color[1] current_color[2] current_color[3]

        materials_group.Controls.Add material_listview

        local filter_label = dotNetObject "System.Windows.Forms.Label"
        filter_label.name = "filter_label"
        filter_label.text = "Filter:"
        filter_label.Location = dotNetObject "System.Drawing.Point" 10 278
        filter_label.AutoSize = true
        filter_label.Anchor = dotnet.combineEnums filter_label.anchor.Left filter_label.anchor.Bottom

        materials_group.controls.add filter_label

        material_filter_text = dotNetObject "System.Windows.Forms.TextBox"
        material_filter_text.name = "material_filter_text"
        material_filter_text.Location = dotNetObject "System.Drawing.Point" 50 275
        material_filter_text.Size = dotNetObject "System.Drawing.Size" 100 20
        material_filter_text.Anchor = dotnet.combineEnums material_filter_text.anchor.Bottom material_filter_text.anchor.Left material_filter_text.anchor.Right
        dotNet.setLifetimeControl material_filter_text #dotnet
        dotNet.addEventHandler material_filter_text "KeyUp" material_filter_text_changed

        materials_group.controls.add material_filter_text

        local update_list_button = dotNetObject "System.Windows.Forms.Button"
        update_list_button.name = "update_list_button"
        update_list_button.text = "Update List"
        update_list_button.Location = dotNetObject "System.Drawing.Point" 215 275
        update_list_button.Size = dotNetObject "System.Drawing.Size" 75 25
        update_list_button.flatStyle = update_list_button.flatStyle.system
        update_list_button.Anchor = dotnet.combineEnums update_list_button.anchor.Right update_list_button.anchor.Bottom
        dotNet.setLifetimeControl update_list_button #dotnet
        dotNet.addEventHandler update_list_button "Click" update_list_button_pressed

        materials_group.controls.add update_list_button

        /********************************************************************
        Options Group
        ********************************************************************/
        local option_group = dotNetObject "System.Windows.Forms.GroupBox"
        option_group.name = "option_group"
        option_group.text = "Options"
        option_group.Location = dotNetObject "System.Drawing.Point" 5 315
        option_group.Size = dotNetObject "System.Drawing.Size" 300 65
        option_group.flatStyle = option_group.flatStyle.system
        option_group.Anchor = dotnet.combineEnums option_group.anchor.Left option_group.anchor.Right option_group.anchor.Bottom
        option_group.suspendLayout()

        material_list_form.controls.add option_group

        show_sub_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        show_sub_checkbox.Name = "show_sub_checkbox"
        show_sub_checkbox.Text = "Show Sub Materials"
        show_sub_checkbox.Location = dotNetObject "System.Drawing.Point" 10 20
        show_sub_checkbox.AutoSize = True
        dotNet.setLifetimeControl show_sub_checkbox #dotnet
        dotNet.addEventHandler show_sub_checkbox "CheckedChanged" show_sub_checkbox_changed

        option_group.controls.add show_sub_checkbox

        force_update_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        force_update_checkbox.Name = "force_update_checkbox"
        force_update_checkbox.Text = "Force Material Update"
        force_update_checkbox.Location = dotNetObject "System.Drawing.Point" 10 43
        force_update_checkbox.AutoSize = True

        option_group.controls.add force_update_checkbox

        /********************************************************************
        Material Editor Group
        ********************************************************************/
        local editor_group = dotNetObject "System.Windows.Forms.GroupBox"
        editor_group.name = "editor_group"
        editor_group.text = "Material Editor"
        editor_group.Location = dotNetObject "System.Drawing.Point" 5 385
        editor_group.Size = dotNetObject "System.Drawing.Size" 300 80
        editor_group.flatStyle = editor_group.flatStyle.system
        editor_group.Anchor = dotnet.combineEnums editor_group.anchor.Left editor_group.anchor.Right editor_group.anchor.Bottom
        editor_group.suspendLayout()

        material_list_form.controls.add editor_group

        local show_material_button = dotNetObject "System.Windows.Forms.Button"
        show_material_button.name = "show_material_button"
        show_material_button.text = "Show Material(s) in Editor"
        show_material_button.Location = dotNetObject "System.Drawing.Point" 10 20
        show_material_button.Size = dotNetObject "System.Drawing.Size" 180 25
        show_material_button.flatStyle = show_material_button.flatStyle.system
        show_material_button.Anchor = dotnet.combineEnums show_material_button.anchor.Left show_material_button.anchor.Right
        dotNet.setLifetimeControl show_material_button #dotnet
        dotNet.addEventHandler show_material_button "Click" show_material_button_pressed

        editor_group.controls.add show_material_button

        local slot_label = dotNetObject "System.Windows.Forms.Label"
        slot_label.name = "slot_label"
        slot_label.text = "Slot:"
        slot_label.Location = dotNetObject "System.Drawing.Point" 200 25
        slot_label.AutoSize = true
        slot_label.Anchor = dotnet.combineEnums slot_label.anchor.Right

        editor_group.controls.add slot_label

        slot_spinner = dotNetObject "System.Windows.Forms.NumericUpDown"
        slot_spinner.name = "slot_spinner"
        slot_spinner.Location = dotNetObject "System.Drawing.Point" 230 22
        slot_spinner.Size = dotNetObject "System.Drawing.Size" 40 20
        slot_spinner.Minimum = 1
        slot_spinner.Maximum = 24
        slot_spinner.Value = 1
        slot_spinner.Anchor = slot_spinner.anchor.Right
        dotNet.setLifetimeControl slot_spinner #dotnet
        dotNet.addEventHandler slot_spinner "MouseDown" spinner_mouse_click

        editor_group.controls.add slot_spinner

        local reset_editor_button = dotNetObject "System.Windows.Forms.Button"
        reset_editor_button.name = "reset_editor_button"
        reset_editor_button.text = "Reset Editor"
        reset_editor_button.Location = dotNetObject "System.Drawing.Point" 10 50
        reset_editor_button.Size = dotNetObject "System.Drawing.Size" 180 25
        reset_editor_button.flatStyle = reset_editor_button.flatStyle.system
        reset_editor_button.Anchor = dotnet.combineEnums reset_editor_button.anchor.Left reset_editor_button.anchor.Right
        dotNet.setLifetimeControl reset_editor_button #dotnet
        dotNet.addEventHandler reset_editor_button "Click" reset_editor_button_pressed

        editor_group.controls.add reset_editor_button

        defalut_material_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        defalut_material_checkbox.Name = "defalut_material_checkbox"
        defalut_material_checkbox.Text = "Std Materials"
        defalut_material_checkbox.Location = dotNetObject "System.Drawing.Point" 185 53
        defalut_material_checkbox.AutoSize = true
        defalut_material_checkbox.Checked = true
        defalut_material_checkbox.Anchor = defalut_material_checkbox.anchor.Right

        editor_group.controls.add defalut_material_checkbox

        /********************************************************************
        Select Group
        ********************************************************************/
        local select_group = dotNetObject "System.Windows.Forms.GroupBox"
        select_group.name = "select_group"
        select_group.text = "Selection"
        select_group.Location = dotNetObject "System.Drawing.Point" 5 470
        select_group.Size = dotNetObject "System.Drawing.Size" 300 110
        select_group.flatStyle = select_group.flatStyle.system
        select_group.Anchor = dotnet.combineEnums select_group.anchor.Left select_group.anchor.Right select_group.anchor.Bottom
        select_group.suspendLayout()

        material_list_form.controls.add select_group

        local select_objects_button = dotNetObject "System.Windows.Forms.Button"
        select_objects_button.name = "select_objects_button"
        select_objects_button.text = "Select Objects with Material(s)"
        select_objects_button.Location = dotNetObject "System.Drawing.Point" 10 20
        select_objects_button.Size = dotNetObject "System.Drawing.Size" 170 25
        select_objects_button.flatStyle = select_objects_button.flatStyle.system
        select_objects_button.Anchor = dotnet.combineEnums select_objects_button.anchor.Left select_objects_button.anchor.Right
        dotNet.setLifetimeControl select_objects_button #dotnet
        dotNet.addEventHandler select_objects_button "Click" select_objects_button_pressed

        select_group.controls.add select_objects_button

        add_selection_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        add_selection_checkbox.Name = "add_selection_checkbox"
        add_selection_checkbox.Text = "Keep Selected"
        add_selection_checkbox.Location = dotNetObject "System.Drawing.Point" 185 23
        add_selection_checkbox.AutoSize = true
        add_selection_checkbox.Anchor = add_selection_checkbox.anchor.Right

        select_group.controls.add add_selection_checkbox

        local assign_material_button = dotNetObject "System.Windows.Forms.Button"
        assign_material_button.name = "assign_material_button"
        assign_material_button.text = "Assign Material to Selected Objects"
        assign_material_button.Location = dotNetObject "System.Drawing.Point" 10 50
        assign_material_button.Size = dotNetObject "System.Drawing.Size" 280 25
        assign_material_button.flatStyle = assign_material_button.flatStyle.system
        assign_material_button.Anchor = dotnet.combineEnums assign_material_button.anchor.Left assign_material_button.anchor.Right
        dotNet.setLifetimeControl assign_material_button #dotnet
        dotNet.addEventHandler assign_material_button "Click" assign_material_button_pressed

        select_group.controls.add assign_material_button

        local show_obj_mat_button = dotNetObject "System.Windows.Forms.Button"
        show_obj_mat_button.name = "show_obj_mat_button"
        show_obj_mat_button.text = "Show Materials of Selected Objects"
        show_obj_mat_button.Location = dotNetObject "System.Drawing.Point" 10 80
        show_obj_mat_button.Size = dotNetObject "System.Drawing.Size" 280 25
        show_obj_mat_button.flatStyle = show_obj_mat_button.flatStyle.system
        show_obj_mat_button.Anchor = dotnet.combineEnums show_obj_mat_button.anchor.Left show_obj_mat_button.anchor.Right
        dotNet.setLifetimeControl show_obj_mat_button #dotnet
        dotNet.addEventHandler show_obj_mat_button "Click" show_obj_mat_button_pressed

        select_group.controls.add show_obj_mat_button

        /********************************************************************
        ID Group
        ********************************************************************/
        local id_group = dotNetObject "System.Windows.Forms.GroupBox"
        id_group.name = "id_group"
        id_group.text = "IDs"
        id_group.Location = dotNetObject "System.Drawing.Point" 5 585
        id_group.Size = dotNetObject "System.Drawing.Size" 300 80
        id_group.flatStyle = id_group.flatStyle.system
        id_group.Anchor = dotnet.combineEnums id_group.anchor.Left id_group.anchor.Right id_group.anchor.Bottom
        id_group.suspendLayout()

        material_list_form.controls.add id_group

        local obj_id_button = dotNetObject "System.Windows.Forms.Button"
        obj_id_button.name = "obj_id_button"
        obj_id_button.text = "Assign Objects with Material(s):"
        obj_id_button.Location = dotNetObject "System.Drawing.Point" 10 20
        obj_id_button.Size = dotNetObject "System.Drawing.Size" 200 25
        obj_id_button.flatStyle = obj_id_button.flatStyle.system
        obj_id_button.Anchor = dotnet.combineEnums obj_id_button.anchor.Left obj_id_button.anchor.Right
        dotNet.setLifetimeControl obj_id_button #dotnet
        dotNet.addEventHandler obj_id_button "Click" obj_id_button_pressed

        id_group.controls.add obj_id_button

        obj_id_spinner = dotNetObject "System.Windows.Forms.NumericUpDown"
        obj_id_spinner.name = "obj_id_spinner"
        obj_id_spinner.Location = dotNetObject "System.Drawing.Point" 220 22
        obj_id_spinner.Size = dotNetObject "System.Drawing.Size" 60 20
        obj_id_spinner.Minimum = 0
        obj_id_spinner.Maximum = 65535
        obj_id_spinner.Value = 1
        obj_id_spinner.Anchor = obj_id_spinner.anchor.Right
        dotNet.setLifetimeControl obj_id_spinner #dotnet
        dotNet.addEventHandler obj_id_spinner "MouseDown" spinner_mouse_click

        id_group.controls.add obj_id_spinner

        local mat_id_button = dotNetObject "System.Windows.Forms.Button"
        mat_id_button.name = "mat_id_button"
        mat_id_button.text = "Assign Material(s):"
        mat_id_button.Location = dotNetObject "System.Drawing.Point" 10 50
        mat_id_button.Size = dotNetObject "System.Drawing.Size" 200 25
        mat_id_button.flatStyle = mat_id_button.flatStyle.system
        mat_id_button.Anchor = dotnet.combineEnums mat_id_button.anchor.Left mat_id_button.anchor.Right
        dotNet.setLifetimeControl mat_id_button #dotnet
        dotNet.addEventHandler mat_id_button "Click" mat_id_button_pressed

        id_group.controls.add mat_id_button

        mat_id_spinner = dotNetObject "System.Windows.Forms.NumericUpDown"
        mat_id_spinner.name = "mat_id_spinner"
        mat_id_spinner.Location = dotNetObject "System.Drawing.Point" 220 52
        mat_id_spinner.Size = dotNetObject "System.Drawing.Size" 40 20
        mat_id_spinner.Minimum = 0
        mat_id_spinner.Maximum = 15
        mat_id_spinner.Value = 1
        mat_id_spinner.Anchor = mat_id_spinner.anchor.Right
        dotNet.setLifetimeControl mat_id_spinner #dotnet
        dotNet.addEventHandler mat_id_spinner "MouseDown" spinner_mouse_click

        id_group.controls.add mat_id_spinner

        /********************************************************************
        Misc Group
        ********************************************************************/
        local misc_group = dotNetObject "System.Windows.Forms.GroupBox"
        misc_group.name = "misc_group"
        misc_group.text = "Misc"
        misc_group.Location = dotNetObject "System.Drawing.Point" 5 670
        misc_group.Size = dotNetObject "System.Drawing.Size" 300 80
        misc_group.flatStyle = misc_group.flatStyle.system
        misc_group.Anchor = dotnet.combineEnums misc_group.anchor.Left misc_group.anchor.Right misc_group.anchor.Bottom
        misc_group.suspendLayout()

        material_list_form.controls.add misc_group

        local save_materials_button = dotNetObject "System.Windows.Forms.Button"
        save_materials_button.name = "save_materials_button"
        save_materials_button.text = "Save Selected Materials to File"
        save_materials_button.Location = dotNetObject "System.Drawing.Point" 10 20
        save_materials_button.Size = dotNetObject "System.Drawing.Size" 280 25
        save_materials_button.flatStyle = save_materials_button.flatStyle.system
        save_materials_button.Anchor = dotnet.combineEnums save_materials_button.anchor.Left save_materials_button.anchor.Right
        dotNet.setLifetimeControl save_materials_button #dotnet
        dotNet.addEventHandler save_materials_button "Click" save_materials_button_pressed

        misc_group.controls.add save_materials_button

        local instance_duplicate_button = dotNetObject "System.Windows.Forms.Button"
        instance_duplicate_button.name = "instance_duplicate_button"
        instance_duplicate_button.text = "Instance Duplicate Name Materials"
        instance_duplicate_button.Location = dotNetObject "System.Drawing.Point" 10 50
        instance_duplicate_button.Size = dotNetObject "System.Drawing.Size" 280 25
        instance_duplicate_button.flatStyle = instance_duplicate_button.flatStyle.system
        instance_duplicate_button.Anchor = dotnet.combineEnums instance_duplicate_button.anchor.Left instance_duplicate_button.anchor.Right
        dotNet.setLifetimeControl instance_duplicate_button #dotnet
        dotNet.addEventHandler instance_duplicate_button "Click" instance_duplicate_button_pressed

        misc_group.controls.add instance_duplicate_button

        /********************************************************************
        Status bar
        ********************************************************************/
        local status_bar = dotNetObject "System.Windows.Forms.StatusStrip"
        status_bar.name = "status_bar"
        status_bar.Location = dotNetObject "System.Drawing.Point" 0 0
        status_bar.dock = status_bar.dock.bottom
        status_bar.sizingGrip = false

        material_list_form.controls.add status_bar

        --The status bar color if not inherted from the max form
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #window) * 255
        status_bar.backColor = dn_color current_color[1] current_color[2] current_color[3]

        status_label = dotNetObject "System.Windows.Forms.ToolStripStatusLabel"
        status_label.name = "status_label"
        status_label.text = "INFO: Standing by..."
        status_label.autoSize = true

        status_bar.items.add status_label

        /********************************************************************
        Finalize UI
        ********************************************************************/
        misc_group.resumeLayout()
        id_group.resumeLayout()
        select_group.resumeLayout()
        editor_group.resumeLayout()
        option_group.resumeLayout()
        materials_group.resumeLayout()
        material_list_form.resumeLayout()

        /********************************************************************
        Refresh and Open UI
        ********************************************************************/
        refresh_material_list()

        form_position = execute (getIniSetting material_list_ini "Dialog" "Position")
        form_size = execute (getIniSetting material_list_ini "Dialog" "Size")

        if form_position == OK do
            form_position = point2 100 100
        if form_size == OK do
            form_size = point2 325 820

        if form_position.x > sysinfo.desktopsize.x - form_size.x do
            form_position.x = sysinfo.desktopsize.x - form_size.x
        if form_position.y > sysinfo.desktopsize.y - (form_size.y + 60) do
            form_position.y = sysinfo.desktopsize.y - (form_size.y + 60)
        if form_position.x < 0 do
            form_position.x = 0
        if form_position.y < 0 do
            form_position.y = 0

        material_list_form.startPosition = material_list_form.startPosition.manual
        material_list_form.location = dotNetObject "System.Drawing.Point" form_position.x form_position.y
        material_list_form.size = dotNetObject "System.Drawing.Size" form_size.x form_size.y

        if material_listview.width > 280 do
            (material_listview.Columns.Item 0).Width = material_listview.width - 115

        local the_pointer = DotNetObject "System.IntPtr" (windows.getMAXHWND())
        --CRITICAL! Do not use Win32HandleWrapper, it was removed in >= 2015
        --local the_handle = DotNetObject "MaxCustomControls.Win32HandleWrapper" the_pointer
        local the_handle = (dotnetclass "System.Windows.Forms.Control").fromHandle the_pointer

        material_list_form.Show(the_handle)
    ) --end fn

    --CRITICAL! The following is required since .net/mxs can crash building a UI
    attempts = 1
    while attempts < 10 do
    (
        try
        (
            setup_ui()
            attempts = 100
        ) --end try
        catch
        (
            print("WARNING: Failed to launch Material List. Attempt: " + (attempts as string))
            attempts += 1
        ) --end catch
    ) --end while

    if attempts == 10 do
        messageBox "ERROR: Could not launch Material List" title:"Launch Error"
) --end scope
